//
//------------------------------------------------------------------------------
// Name    : ivt.bas
// Author  : Jerry Messina
// Date    : 12 MAY 2023
// Version : 2.20
// Notes   : IVT routines for xv18 VIC vectored interrupt mode
//
// ver 2.20  12 MAY 2023
//  - add saving TABLAT register as part of SAVE_CONTEXT()/RESTORE_CONTEXT()
//      this addresses the issue of getting interrupted between the TBLRD and TABLAT instructions
//  - add error check for v2.0 format '#option _IVT_SAVE_CONTEXT=true,false'
// ver 2.10  5/29/2022
//  - change '#option _IVT_SAVE_CONTEXT' to specify the required level of context 
//      support based on the priorities used and exclude unneeded code/data structures
//  - add macros SAVE_SYSTEM_CONTEXT()/RESTORE_SYSTEM_CONTEXT()
//      these allow multiple ISR handlers to save the SF system variables context
//      in shared memory and replace the built-in 'save(0)/restore' functions
//  - add '#option _IVT_DEFAULT_IPEN' to allow setting IPEN at startup
//  - update interrupt handler template
// ver 2.00  2/15/2021
//  - restructure vector table usage...
//      simpler, less code space, and faster
//  - remove intermediate jump table as it's no longer required
//      ENTER_INTR_HANDLER() adds any required alignment so we can 
//      directly set IVT without using a jump table
//  - change CREATE_IVT/SET_IVT_VECTOR/END_IVT macros to track vectors
//      using __intvector#v(ivt_no)
//  - unused vectors are now set to use default_interrupt_handler() 
//      automatically by CREATE_IVT and no longer require being
//      specified by the user using SET_IVT_HANDLER
//  - now supports up to 256 vectors, but currently the max is set
//      to 128 (requires adding more PIEx/PIRx definitions)
//  - rename BEGIN_REGISTER_INTR_HANDLER/END_REGISTER_INTR_HANDLER
//      macros to BEGIN_SB_COMMENT_BLOCK/END_SB_COMMENT_BLOCK to
//      better reflect their usage
//  - add '#if (_NUM_INT_VECTORS > 80)' conditional for IPR/PIE/PIR
//      to account for K83 new '#define _intvectors = 79' (datasheet error)
//  - add ENABLE_INTERRUPT()/DISABLE_INTERRUPT() for high-priority use
//  - add SAVE_CONTEXT/SAVE_CONTEXT(priority)/SAVE_CONTEXT_ISS
//      and RESTORE_CONTEXT forms for different priority schemes
//  - add INIT_VIC()
// ver 1.30
//  - add comments about using save(0)/restore and SAVE_CONTEXT
//    revise interrupt handler template to show usage
//  - add public HIGH_PRIORITY and LOW_PRIORITY const definitions
// ver 1.20
//  - add '#option _IVT_IVTLOCKED' to specify IVTLOCKED bit location
// ver 1.10
//  - add support for device file '_accessbank'
//      ENTER_INTR_HANDLER()
//      SET_IVTBASE()
//  - add REGISTER_INTR_HANDLER(default_interrupt_handler)
//    to prevent undefined reference to the default handler  
//  - add SAVE_CONTEXT()/RESTORE_CONTEXT() to save TABLEPTR regs
// ver 1.00 9/30/2019
//  - initial version
//------------------------------------------------------------------------------
//
module ivt

// check if device supports vectored interrupts
#if not defined(_xv18) or (_xv18 = 0)       // from device .bas file
  #error _device + " does not support VIC mode"
#endif  

include "system.bas"
// check to make sure we pulled in system.bas with support for '#option org_ram'
#if not defined(SF_SYSTEM_VERSION) or (SF_SYSTEM_VERSION < 2231)
  #error "incorrect SF_SYSTEM_VERSION detected (system.bas)"
#endif  
// number of system variables (requires system.bas 2236 or later)
#if defined(SF_SYSTEM_VERSION) and (SF_SYSTEM_VERSION >= 2236)
  public const SB_SV_VARS = sysmodule.SB_SV_VARS
#else   // assume 25 system vars
  public const SB_SV_VARS = 25
#endif
// alias system.bas SB_SV0 so that it's visible to macros
public dim SB_SV0 as sysmodule.SB_SV0

// number of interrupt vectors
// this setting controls the number of entries in the IVT vector table
// the default setting is taken from the device file '#define _intvectors' value
// any undefined vectors will be set to point to default_interrupt_handler()
// to reduce the IVT table size you can set #option _NUM_INT_VECTORS = highest IRQ number used + 1
#option _NUM_INT_VECTORS = _intvectors
#if not (_NUM_INT_VECTORS in (1 to _intvectors))
  #error _NUM_INT_VECTORS, "invalid _NUM_INT_VECTORS setting... must be 1 to '#define _intvectors'" 
#endif
#if (_NUM_INT_VECTORS > 128)
  #error _NUM_INT_VECTORS, "_NUM_INT_VECTORS currently limited to 128" 
#endif
public const NUM_INT_VECTORS = _NUM_INT_VECTORS

// this module defines a default_interrupt_handler() routine that you can use for all
// unused interrupts. set this option false if you want to use your own routine
#option _IVT_DEFAULT_HANDLER = true

// default Interrupt Priority Enable setting
// sets IPEN value at startup
//  = 0  priority disabled... high-priority only (default)
//  = 1  high and low priorities enabled
#option _IVT_DEFAULT_IPEN = 0
#if not(_IVT_DEFAULT_IPEN in (0, 1))
  #error _IVT_DEFAULT_IPEN, "invalid IPEN setting"
#endif
   
// SAVE_CONTEXT support (_IVT_SAVE_CONTEXT v2.10)
// macros and data structures for saving/restoring the TABLEPTR registers and SF system variables
// (see 'interrupt context' discussion below for details)
// the #option setting is a bit-field value and may have multiple bits set, allowing
// you to tailor the level of context support to that required by the application
// ie '#option _IVT_SAVE_CONTEXT = %0101    // low priority TBLPTR + high-priority TBLPTR
//
// SAVE_CONTEXT bit field values
#define _IVTscNONE  = %0000         // no context support (default setting)
#define _IVTscHPTBL = %0001         // ---1 high-priority TBLPTR
#define _IVTscHPSYS = %0010         // --1- high-priority SF SYSTEM variables
#define _IVTscLPTBL = %0100         // -1-- low priority TBLPTR (requires IPEN = 1) 
#define _IVTscLPSYS = %1000         // 1--- low-priority SF SYSTEM variables (requires IPEN = 1)
#define _IVTscHP    = %0011         // high-priority TBLPTR + SYSTEM
#define _IVTscLP    = %1100         // low-priority TBLPTR + SYSTEM
#define _IVTscALL   = %1111         // high and low priority TBLPTR + SYSTEM
public const
    IVTscNONE  = _IVTscNONE,
    IVTscHPTBL = _IVTscHPTBL,
    IVTscHPSYS = _IVTscHPSYS,
    IVTscLPTBL = _IVTscLPTBL,
    IVTscLPSYS = _IVTscLPSYS,
    IVTscHP    = _IVTscHP,
    IVTscLP    = _IVTscLP,
    IVTscALL   = _IVTscALL

// check for previous version option setting
#if (isoption(_IVT_SAVE_CONTEXT) and _IVT_SAVE_CONTEXT in (true, false))
  #error _IVT_SAVE_CONTEXT, "_IVT_SAVE_CONTEXT option setting invalid for IVT v2.10"
#endif
#option _IVT_SAVE_CONTEXT = %0000       // _IVTscNONE
#if not(_IVT_SAVE_CONTEXT in (0 to 15))
  #error _IVT_SAVE_CONTEXT, "_IVT_SAVE_CONTEXT invalid"
#endif
public const ivt_save_context = _IVT_SAVE_CONTEXT
    
// automatically call INIT_VIC() at startup to set the VIC peripheral poweron settings
#option _IVT_INIT_VIC = false

// automatically register the interrupt handlers as part of SET_IVT_HANDLER
// if this option is false you MUST use REGISTER_INTR_HANDLER() to declare
// the handlers before they're referenced, so leave this set true
#option _IVT_AUTO_REGISTER_HANDLERS = true

// IVTLOCKED bit (default, override if required)
#option _IVTLOCKED = IVTLOCK.0
// split _IVTLOCKED into register and bit for asm routines
public dim IVTLOCK_REG as _IVTLOCKED
public const IVTLOCKED = _IVTLOCKED@

// enable multivector mode config setting
config MVECEN = ON

// the IVT1WAY configuration bit determines the number of times the IVTLOCKED bit
// can be cleared and set after a device reset
// with IVT1WAY = ON, the IVTLOCKED bit can be manipulated only once
// IVT1WAY defaults to ON, so if you want to be able to change tables at runtime 
// it must be set OFF
config IVT1WAY = OFF

// get access bank no from the device file '#define _accessbank'
public const ACCESSBANK = _accessbank

// priority level definitions
public const
    HIGH_PRIORITY = 1,
    LOW_PRIORITY = 0
    
// INTCON register interrupt enable bits
public const
    GIE = 7,
    GIEL = 6
    
// values used by CheckParam()
public const 
    etError = 0,
    cpConst = $01

//
//-----------------------------------------------------------------------------
// ENABLE_INTR_PRIORITY()
// DISABLE_INTR_PRIORITY()
//   sets IPEN intr priority enable bit
// default setting is IPEN = 0 (priority disabled... all interrupts are high-priority)
//-----------------------------------------------------------------------------
//
public inline sub ENABLE_INTR_PRIORITY()
    IPEN = 1
end sub

public inline sub DISABLE_INTR_PRIORITY()
    IPEN = 0    // default setting
end sub

//
//-----------------------------------------------------------------------------
// ENABLE_INTERRUPT()
// DISABLE_INTERRUPT()
// ENABLE_INTERRUPT(priority)
// DISABLE_INTERRUPT(priority)
//   sets GIE/GIEL global interrupt enable bits
// parameters:
//   priority = 0    enable/disable low-priority intr (GIEL)
//   priority = 1    enable/disable high-priority intr (GIE)
// note: setting GIE=0 disables all interrupts (both levels)
//-----------------------------------------------------------------------------
//
// for use with priority disabled (IPEN=0, high-priority only)
public macro ENABLE_INTERRUPT()
    INTCON0.bits(GIE) = 1
end macro

public macro DISABLE_INTERRUPT()
    INTCON0.bits(GIE) = 0
end macro

// general use/priority enabled
public macro ENABLE_INTERRUPT(priority)
    if (priority = 0) then
        INTCON0.bits(GIEL) = 1
    else
        INTCON0.bits(GIE) = 1
    endif        
end macro

public macro DISABLE_INTERRUPT(priority)
    if (priority = 0) then
        INTCON0.bits(GIEL) = 0
    else
        INTCON0.bits(GIE) = 0
    endif        
end macro

// get current GIE/GIEL register setting
public inline function INTERRUPT_ENABLE() as INTCON0
end function

//
//-----------------------------------------------------------------------------
// CREATE_IVT(table_name)
//   create an IVT table
// parameters:
//   table_name =  name of table
// set all IVT entry variables to point to the default_interrupt_handler
// after creating the table, change the entries using SET_IVT_HANDLER
// the actual IVT will be built by END_IVT()
// note: this is done in blocks of 48 due to the mpasm limit on 100 
//   statements between .while/.endw (and this gets expanded in the .asm file)
//   same with END_IVT()
//-----------------------------------------------------------------------------
//
public macro CREATE_IVT(table_name)
  asm-
  ivt_no set 0
  ; vectors 0-47
    last_ivt set NUM_INT_VECTORS
  .if (NUM_INT_VECTORS >= 48)
    last_ivt set 48
  .endif        
  .while (ivt_no < last_ivt)
      __intvector#v(ivt_no) set default_interrupt_handler
      ivt_no++
  .endw      
  ; vectors 48-95
    last_ivt set NUM_INT_VECTORS
  .if (NUM_INT_VECTORS >= 48*2)
    last_ivt set 48*2
  .endif        
  .while (ivt_no < last_ivt)
      __intvector#v(ivt_no) set default_interrupt_handler
      ivt_no++
  .endw      
  ; vectors 96-143
    last_ivt set NUM_INT_VECTORS
  .if (NUM_INT_VECTORS >= 48*3)
    last_ivt set 48*3
  .endif        
  .while (ivt_no < last_ivt)
      __intvector#v(ivt_no) set default_interrupt_handler
      ivt_no++
  .endw      
  ; vectors 144-191
    last_ivt set NUM_INT_VECTORS
  .if (NUM_INT_VECTORS >= 48*4)
    last_ivt set 48*4
  .endif        
  .while (ivt_no < last_ivt)
      __intvector#v(ivt_no) set default_interrupt_handler
      ivt_no++
  .endw      
  ; vectors 192-239
    last_ivt set NUM_INT_VECTORS
  .if (NUM_INT_VECTORS >= 48*5)
    last_ivt set 48*5
  .endif        
  .while (ivt_no < last_ivt)
      __intvector#v(ivt_no) set default_interrupt_handler
      ivt_no++
  .endw      
  ; vectors 240-255
  .if (NUM_INT_VECTORS > 240)
    last_ivt set NUM_INT_VECTORS
  .endif        
  .while (ivt_no < last_ivt)
      __intvector#v(ivt_no) set default_interrupt_handler
      ivt_no++
  .endw      

  ; align to even address (which we should be anyway)
  org (($+1) & 0x3FFFE)
  
  ; add a jump over ivt table to make sure we never try to 
  ; execute the ivt as instructions
  goto $+4 + (NUM_INT_VECTORS*2)
  
  ; set start of ivt vector table
  table_name org $
  end asm
end macro

//
//-----------------------------------------------------------------------------
// END_IVT()
//   build an IVT table defined using the CREATE_IVT/SET_IVT_HANDLER macros
//   this creates the actual vector table in program memory space
// note: the table holds 16-bit word addresses set using 'addr >> 2'
//   to match the vic limitation used to access the full 256K memory space   
//-----------------------------------------------------------------------------
//
public macro END_IVT()
  asm-
  ; create ivt in program memory
  vshift set 2    ; handler addresses get divided by 4 (shift right >> 2)
  valign set 2    ; account for isr NOP alignment
  ivt_no set 0
  ; vectors 0-47
    last_ivt set NUM_INT_VECTORS
  .if (NUM_INT_VECTORS >= 48)
    last_ivt set 48
  .endif        
  .while (ivt_no < last_ivt)
      DW (__intvector#v(ivt_no) + valign) >> vshift
      ivt_no++
  .endw      
  ; vectors 48-95
    last_ivt set NUM_INT_VECTORS
  .if (NUM_INT_VECTORS >= 48*2)
    last_ivt set 48*2
  .endif        
  .while (ivt_no < last_ivt)
      DW (__intvector#v(ivt_no) + valign) >> vshift
      ivt_no++
  .endw      
  ; vectors 96-143
    last_ivt set NUM_INT_VECTORS
  .if (NUM_INT_VECTORS >= 48*3)
    last_ivt set 48*3
  .endif        
  .while (ivt_no < last_ivt)
      DW (__intvector#v(ivt_no) + valign) >> vshift
      ivt_no++
  .endw      
  ; vectors 144-191
    last_ivt set NUM_INT_VECTORS
  .if (NUM_INT_VECTORS >= 48*4)
    last_ivt set 48*4
  .endif        
  .while (ivt_no < last_ivt)
      DW (__intvector#v(ivt_no) + valign) >> vshift
      ivt_no++
  .endw      
  ; vectors 192-239
    last_ivt set NUM_INT_VECTORS
  .if (NUM_INT_VECTORS >= 48*5)
    last_ivt set 48*5
  .endif        
  .while (ivt_no < last_ivt)
      DW (__intvector#v(ivt_no) + valign) >> vshift
      ivt_no++
  .endw      
  ; vectors 240-255
  .if (NUM_INT_VECTORS > 240)
    last_ivt set NUM_INT_VECTORS
  .endif        
  .while (ivt_no < last_ivt)
      DW (__intvector#v(ivt_no) + valign) >> vshift
      ivt_no++
  .endw      
  end asm
end macro

//
//-----------------------------------------------------------------------------
// SET_IVT_HANDLER(vector_no, handler)
//   add an interrupt handler to the current vector table (CREATE_IVT)
// parameters:
//   vector_no = 0 to NUM_INT_VECTORS-1 (can use 'IRQ_xxx' definitions from the device file)
//   handler = name of the interrupt handler (must be an 'event' routine)
// this will be used to set the vector table entry by END_IVT()
//-----------------------------------------------------------------------------
//
public macro SET_IVT_HANDLER(vector_no, handler)
    if ((vector_no) > (NUM_INT_VECTORS-1)) then
        checkparam(etError, "invalid vector_no parameter: verify _NUM_INT_VECTORS")
    endif
  // set the vector tracking variable to the event handler routine
  // this will be used when the vector table is built by END_IVT 
  asm-
  __intvector#v(vector_no) set handler
  end asm
  // create a reference to the event handler to satisfy that it's used...
  // the unused code marked with SB_COMMENT_BLOCK_BEGIN/SB_COMMENT_BLOCK_END
  // will be removed by the ICC
#if (_IVT_AUTO_REGISTER_HANDLERS)
  asm-
SB_COMMENT_BLOCK_BEGIN:
  end asm  
  addressof(handler)
  asm-
SB_COMMENT_BLOCK_END:
  end asm  
#endif
end macro

//
//-----------------------------------------------------------------------------
// BEGIN_SB_COMMENT_BLOCK()
// END_SB_COMMENT_BLOCK()
//   these macros can be used to mark a section of code with SB_COMMENT_BLOCK_BEGIN
//   and SB_COMMENT_BLOCK_END markers, which will allow the ICC code generator to 
//   comment out and remove the code between BLOCK_BEGIN and BLOCK_END
// for internal use
//-----------------------------------------------------------------------------
//
private macro BEGIN_SB_COMMENT_BLOCK()
    asm-
SB_COMMENT_BLOCK_BEGIN:
    end asm  
end macro

private macro END_SB_COMMENT_BLOCK()
    asm-
SB_COMMENT_BLOCK_END:
    end asm  
end macro

//
//-----------------------------------------------------------------------------
// REGISTER_INTR_HANDLER(handler)
//   creates a reference to an interrupt handler event()
//   this is required so that SF recognizes the event is used and does not remove it
// parameters:
//   handler = name of the interrupt handler (must be an 'event' routine)
// note: this macro generates code that is never executed
//   BEGIN_SB_COMMENT_BLOCK()/END_SB_COMMENT_BLOCK() can be used to create a comment 
//   block section which will allow the ICC code generator to remove the unneeded code
// NOTE: 
//   REGISTER_INTR_HANDLER() *MUST* be called for each interrupt handler event() 
//   routine before using the handler with SET_IVT_HANDLER()
//   setting '#option _IVT_AUTO_REGISTER_HANDLERS = true' will do this automatically...
//   otherwise you must add it manually
//-----------------------------------------------------------------------------
//
public macro REGISTER_INTR_HANDLER(handler)
    addressof(handler)
end macro

//
//-----------------------------------------------------------------------------
// SET_IVTBASE(table_name)
//   set IVTBASE U/H/L registers to point to an ivt table
// parameters:
//   table_name = name of IVT table created with CREATE_IVT 
// The IVTBASE register is user programmable and is used to determine the start address 
// of the Interrupt Vector Table, and the IVTLOCK register is used to prevent any 
// unintended writes to the IVTBASE register.
// When the IVT1WAY configuration bit is set, the IVTLOCKED bit can be cleared and set 
// only once after a device reset
// note: this routine assumes that IVTBASEU/H/L, IVTLOCK, and INTCON0 registers
//  are located in the access bank (updated v1.10 - adds banksel for Q43 IVT regs)
// **SET_IVTBASE disables interrupts and leaves them disabled on exit**
//-----------------------------------------------------------------------------
//
public macro SET_IVTBASE(table_name)
  asm-
    ; disable interrupts
    bcf     INTCON0, GIE

    ; save BSR and set it to access the IVTxxxx regs
    ; (MOVFF will get translated to MOVFFL if reqd)
    movff   BSR, PRODL
    banksel IVTLOCK_REG

    ; unlock IVTLOCK register 
    movlw   0x55
    movwf   IVTLOCK_REG
    movlw   0xAA
    movwf   IVTLOCK_REG
    ; clear IVTLOCKED bit to enable writes
    bcf     IVTLOCK_REG, IVTLOCKED

    ; set IVTBASE registers with the address of the IVT table
    movlw   low table_name
    movwf   IVTBASEL
    movlw   high table_name
    movwf   IVTBASEH
    movlw   upper table_name
    movwf   IVTBASEU

    ; unlock IVTLOCK register 
    movlw   0x55
    movwf   IVTLOCK_REG
    movlw   0xAA
    movwf   IVTLOCK_REG
    ; set IVTLOCKED bit to disable writes
    bsf     IVTLOCK_REG, IVTLOCKED

    ; restore BSR to initial setting
    movff   PRODL, BSR
  end asm    
end macro

//
//-----------------------------------------------------------------------------
// INTR_PRIORITY(vector_no, val)
//   sets/clears IPRx interrupt priority register bit
// parameters:
//   vector_no = 0 to NUM_INT_VECTORS-1 (can use 'IRQ_xxx' definitions from the device file)
//   val = priority... 0=low priority, 1=high priority
// default setting = 1 (high priority)
// this routine is not required if IPEN = 0 (priority disabled)
//-----------------------------------------------------------------------------
//
public macro INTR_PRIORITY(vector_no, val)
    if not checkparam(vector_no, cpConst) or not checkparam(val, cpConst) then
        checkparam(etError, "INTR_PRIORITY requires const parameters")
    elseif ((vector_no) > (NUM_INT_VECTORS-1)) then
        checkparam(etError, "invalid vector_no parameter: verify _NUM_INT_VECTORS")
    endif

    if ((vector_no) <= 7) then
        if (val = 1) then
            IPR0 = IPR0 or byte(1 << ((vector_no) mod 8))
        else
            IPR0 = IPR0 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 8) and ((vector_no) <= 15) then
        if (val = 1) then
            IPR1 = IPR1 or byte(1 << ((vector_no) mod 8))
        else
            IPR1 = IPR1 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 16) and ((vector_no) <= 23) then
        if (val = 1) then
            IPR2 = IPR2 or byte(1 << ((vector_no) mod 8))
        else
            IPR2 = IPR2 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 24) and ((vector_no) <= 31) then
        if (val = 1) then
            IPR3 = IPR3 or byte(1 << ((vector_no) mod 8))
        else
            IPR3 = IPR3 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 32) and ((vector_no) <= 39) then
        if (val = 1) then
            IPR4 = IPR4 or byte(1 << ((vector_no) mod 8))
        else
            IPR4 = IPR4 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 40) and ((vector_no) <= 47) then
        if (val = 1) then
            IPR5 = IPR5 or byte(1 << ((vector_no) mod 8))
        else
            IPR5 = IPR5 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 48) and ((vector_no) <= 55) then
        if (val = 1) then
            IPR6 = IPR6 or byte(1 << ((vector_no) mod 8))
        else
            IPR6 = IPR6 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 56) and ((vector_no) <= 63) then
        if (val = 1) then
            IPR7 = IPR7 or byte(1 << ((vector_no) mod 8))
        else
            IPR7 = IPR7 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 64) and ((vector_no) <= 71) then
        if (val = 1) then
            IPR8 = IPR8 or byte(1 << ((vector_no) mod 8))
        else
            IPR8 = IPR8 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 72) and ((vector_no) <= 79) then
        if (val = 1) then
            IPR9 = IPR9 or byte(1 << ((vector_no) mod 8))
        else
            IPR9 = IPR9 and not byte(1 << ((vector_no) mod 8))
        endif
  #if (_NUM_INT_VECTORS > 80)
    elseif ((vector_no) >= 80) and ((vector_no) <= 87) then
        if (val = 1) then
            IPR10 = IPR10 or byte(1 << ((vector_no) mod 8))
        else
            IPR10 = IPR10 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
  #if (_NUM_INT_VECTORS > 88)
    elseif ((vector_no) >= 88) and ((vector_no) <= 95) then
        if (val = 1) then
            IPR11 = IPR11 or byte(1 << ((vector_no) mod 8))
        else
            IPR11 = IPR11 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
  #if (_NUM_INT_VECTORS > 96)
    elseif ((vector_no) >= 96) and ((vector_no) <= 103) then
        if (val = 1) then
            IPR12 = IPR12 or byte(1 << ((vector_no) mod 8))
        else
            IPR12 = IPR12 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
  #if (_NUM_INT_VECTORS > 104)
    elseif ((vector_no) >= 104) and ((vector_no) <= 111) then
        if (val = 1) then
            IPR13 = IPR13 or byte(1 << ((vector_no) mod 8))
        else
            IPR13 = IPR13 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
  #if (_NUM_INT_VECTORS > 112)
    elseif ((vector_no) >= 112) and ((vector_no) <= 119) then
        if (val = 1) then
            IPR14 = IPR14 or byte(1 << ((vector_no) mod 8))
        else
            IPR14 = IPR14 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
  #if (_NUM_INT_VECTORS > 120)
    elseif ((vector_no) >= 120) and ((vector_no) <= 127) then
        if (val = 1) then
            IPR15 = IPR15 or byte(1 << ((vector_no) mod 8))
        else
            IPR15 = IPR15 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
    endif    
end macro

//
//-----------------------------------------------------------------------------
// INTR_ENABLE(vector_no, val)
//   sets/clears PIEx interrupt enable mask register bit
// parameters:
//   vector_no = 0 to NUM_INT_VECTORS-1 (can use 'IRQ_xxx' definitions from the device file)
//   val = enable flag... 0=disabled, 1=enabled
// default setting = 0 (intr disabled)
//-----------------------------------------------------------------------------
//
public macro INTR_ENABLE(vector_no, val)
    if not checkparam(vector_no, cpConst) or not checkparam(val, cpConst) then
        checkparam(etError, "INTR_ENABLE requires const parameters")
    elseif ((vector_no) > (NUM_INT_VECTORS-1)) then
        checkparam(etError, "invalid vector_no parameter: verify _NUM_INT_VECTORS")
    endif

    if ((vector_no) <= 7) then
        if (val = 1) then
            PIE0 = PIE0 or byte(1 << ((vector_no) mod 8))
        else
            PIE0 = PIE0 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 8) and ((vector_no) <= 15) then
        if (val = 1) then
            PIE1 = PIE1 or byte(1 << ((vector_no) mod 8))
        else
            PIE1 = PIE1 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 16) and ((vector_no) <= 23) then
        if (val = 1) then
            PIE2 = PIE2 or byte(1 << ((vector_no) mod 8))
        else
            PIE2 = PIE2 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 24) and ((vector_no) <= 31) then
        if (val = 1) then
            PIE3 = PIE3 or byte(1 << ((vector_no) mod 8))
        else
            PIE3 = PIE3 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 32) and ((vector_no) <= 39) then
        if (val = 1) then
            PIE4 = PIE4 or byte(1 << ((vector_no) mod 8))
        else
            PIE4 = PIE4 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 40) and ((vector_no) <= 47) then
        if (val = 1) then
            PIE5 = PIE5 or byte(1 << ((vector_no) mod 8))
        else
            PIE5 = PIE5 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 48) and ((vector_no) <= 55) then
        if (val = 1) then
            PIE6 = PIE6 or byte(1 << ((vector_no) mod 8))
        else
            PIE6 = PIE6 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 56) and ((vector_no) <= 63) then
        if (val = 1) then
            PIE7 = PIE7 or byte(1 << ((vector_no) mod 8))
        else
            PIE7 = PIE7 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 64) and ((vector_no) <= 71) then
        if (val = 1) then
            PIE8 = PIE8 or byte(1 << ((vector_no) mod 8))
        else
            PIE8 = PIE8 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 72) and ((vector_no) <= 79) then
        if (val = 1) then
            PIE9 = PIE9 or byte(1 << ((vector_no) mod 8))
        else
            PIE9 = PIE9 and not byte(1 << ((vector_no) mod 8))
        endif
  #if (_NUM_INT_VECTORS > 80)
    elseif ((vector_no) >= 80) and ((vector_no) <= 87) then
        if (val = 1) then
            PIE10 = PIE10 or byte(1 << ((vector_no) mod 8))
        else
            PIE10 = PIE10 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
  #if (_NUM_INT_VECTORS > 88)
    elseif ((vector_no) >= 88) and ((vector_no) <= 95) then
        if (val = 1) then
            PIE11 = PIE11 or byte(1 << ((vector_no) mod 8))
        else
            PIE11 = PIE11 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
  #if (_NUM_INT_VECTORS > 96)
    elseif ((vector_no) >= 96) and ((vector_no) <= 103) then
        if (val = 1) then
            PIE12 = PIE12 or byte(1 << ((vector_no) mod 8))
        else
            PIE12 = PIE12 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
  #if (_NUM_INT_VECTORS > 104)
    elseif ((vector_no) >= 104) and ((vector_no) <= 111) then
        if (val = 1) then
            PIE13 = PIE13 or byte(1 << ((vector_no) mod 8))
        else
            PIE13 = PIE13 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
  #if (_NUM_INT_VECTORS > 112)
    elseif ((vector_no) >= 112) and ((vector_no) <= 119) then
        if (val = 1) then
            PIE14 = PIE14 or byte(1 << ((vector_no) mod 8))
        else
            PIE14 = PIE14 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
  #if (_NUM_INT_VECTORS > 120)
    elseif ((vector_no) >= 120) and ((vector_no) <= 127) then
        if (val = 1) then
            PIE15 = PIE15 or byte(1 << ((vector_no) mod 8))
        else
            PIE15 = PIE15 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
    endif    
end macro

//
//-----------------------------------------------------------------------------
// INTR_REQ(vector_no, val)
//   sets/clears PIRx interrupt status flag register bit (IF flag)
// parameters:
//   vector_no = 0 to NUM_INT_VECTORS-1 (can use 'IRQ_xxx' definitions from the device file)
//   val = IF status flag... 0=clear, 1=set
//-----------------------------------------------------------------------------
//
public macro INTR_REQ(vector_no, val)
    if not checkparam(vector_no, cpConst) or not checkparam(val, cpConst) then
        checkparam(etError, "INTR_REQ requires const parameters")
    elseif ((vector_no) > (NUM_INT_VECTORS-1)) then
        checkparam(etError, "invalid vector_no parameter: verify _NUM_INT_VECTORS")
    endif

    if ((vector_no) <= 7) then
        if (val = 1) then
            PIR0 = PIR0 or byte(1 << ((vector_no) mod 8))
        else
            PIR0 = PIR0 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 8) and ((vector_no) <= 15) then
        if (val = 1) then
            PIR1 = PIR1 or byte(1 << ((vector_no) mod 8))
        else
            PIR1 = PIR1 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 16) and ((vector_no) <= 23) then
        if (val = 1) then
            PIR2 = PIR2 or byte(1 << ((vector_no) mod 8))
        else
            PIR2 = PIR2 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 24) and ((vector_no) <= 31) then
        if (val = 1) then
            PIR3 = PIR3 or byte(1 << ((vector_no) mod 8))
        else
            PIR3 = PIR3 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 32) and ((vector_no) <= 39) then
        if (val = 1) then
            PIR4 = PIR4 or byte(1 << ((vector_no) mod 8))
        else
            PIR4 = PIR4 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 40) and ((vector_no) <= 47) then
        if (val = 1) then
            PIR5 = PIR5 or byte(1 << ((vector_no) mod 8))
        else
            PIR5 = PIR5 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 48) and ((vector_no) <= 55) then
        if (val = 1) then
            PIR6 = PIR6 or byte(1 << ((vector_no) mod 8))
        else
            PIR6 = PIR6 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 56) and ((vector_no) <= 63) then
        if (val = 1) then
            PIR7 = PIR7 or byte(1 << ((vector_no) mod 8))
        else
            PIR7 = PIR7 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 64) and ((vector_no) <= 71) then
        if (val = 1) then
            PIR8 = PIR8 or byte(1 << ((vector_no) mod 8))
        else
            PIR8 = PIR8 and not byte(1 << ((vector_no) mod 8))
        endif
    elseif ((vector_no) >= 72) and ((vector_no) <= 79) then
        if (val = 1) then
            PIR9 = PIR9 or byte(1 << ((vector_no) mod 8))
        else
            PIR9 = PIR9 and not byte(1 << ((vector_no) mod 8))
        endif
  #if (_NUM_INT_VECTORS > 80)
    elseif ((vector_no) >= 80) and ((vector_no) <= 87) then
        if (val = 1) then
            PIR10 = PIR10 or byte(1 << ((vector_no) mod 8))
        else
            PIR10 = PIR10 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
  #if (_NUM_INT_VECTORS > 88)
    elseif ((vector_no) >= 88) and ((vector_no) <= 95) then
        if (val = 1) then
            PIR11 = PIR11 or byte(1 << ((vector_no) mod 8))
        else
            PIR11 = PIR11 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
  #if (_NUM_INT_VECTORS > 96)
    elseif ((vector_no) >= 96) and ((vector_no) <= 103) then
        if (val = 1) then
            PIR12 = PIR12 or byte(1 << ((vector_no) mod 8))
        else
            PIR12 = PIR12 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
  #if (_NUM_INT_VECTORS > 104)
    elseif ((vector_no) >= 104) and ((vector_no) <= 111) then
        if (val = 1) then
            PIR13 = PIR13 or byte(1 << ((vector_no) mod 8))
        else
            PIR13 = PIR13 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
  #if (_NUM_INT_VECTORS > 112)
    elseif ((vector_no) >= 112) and ((vector_no) <= 119) then
        if (val = 1) then
            PIR14 = PIR14 or byte(1 << ((vector_no) mod 8))
        else
            PIR14 = PIR14 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
  #if (_NUM_INT_VECTORS > 120)
    elseif ((vector_no) >= 120) and ((vector_no) <= 127) then
        if (val = 1) then
            PIR15 = PIR15 or byte(1 << ((vector_no) mod 8))
        else
            PIR15 = PIR15 and not byte(1 << ((vector_no) mod 8))
        endif
  #endif
    endif    
end macro

//
//-----------------------------------------------------------------------------
// INTR_REQ(vector_no)
//   get PIRx interrupt status flag register contents into WREG (IF masked value)
// parameters:
//   vector_no = 0 to NUM_INT_VECTORS-1 (can use 'IRQ_xxx' definitions from the device file)
// returns:
//   PIRx.IF==0: WREG=0, STATUS.2=1 (Z flag)
//   PIRx.IF<>0: WREG>0, STATUS.2=0 (Z flag)
//-----------------------------------------------------------------------------
//
public macro INTR_REQ(vector_no)
    if not checkparam(vector_no, cpConst) then
        checkparam(etError, "INTR_REQ requires const parameter")
    elseif ((vector_no) > (NUM_INT_VECTORS-1)) then
        checkparam(etError, "invalid vector_no parameter: verify _NUM_INT_VECTORS")
    endif

    if ((vector_no) <= 7) then
        WREG = PIR0 and byte(1 << ((vector_no) mod 8))
    elseif ((vector_no) >= 8) and ((vector_no) <= 15) then
        WREG = PIR1 and byte(1 << ((vector_no) mod 8))
    elseif ((vector_no) >= 16) and ((vector_no) <= 23) then
        WREG = PIR2 and byte(1 << ((vector_no) mod 8))
    elseif ((vector_no) >= 24) and ((vector_no) <= 31) then
        WREG = PIR3 and byte(1 << ((vector_no) mod 8))
    elseif ((vector_no) >= 32) and ((vector_no) <= 39) then
        WREG = PIR4 and byte(1 << ((vector_no) mod 8))
    elseif ((vector_no) >= 40) and ((vector_no) <= 47) then
        WREG = PIR5 and byte(1 << ((vector_no) mod 8))
    elseif ((vector_no) >= 48) and ((vector_no) <= 55) then
        WREG = PIR6 and byte(1 << ((vector_no) mod 8))
    elseif ((vector_no) >= 56) and ((vector_no) <= 63) then
        WREG = PIR7 and byte(1 << ((vector_no) mod 8))
    elseif ((vector_no) >= 64) and ((vector_no) <= 71) then
        WREG = PIR8 and byte(1 << ((vector_no) mod 8))
    elseif ((vector_no) >= 72) and ((vector_no) <= 79) then
        WREG = PIR9 and byte(1 << ((vector_no) mod 8))
  #if (_NUM_INT_VECTORS > 80)
    elseif ((vector_no) >= 80) and ((vector_no) <= 87) then
        WREG = PIR10 and byte(1 << ((vector_no) mod 8))
  #endif
  #if (_NUM_INT_VECTORS > 88)
    elseif ((vector_no) >= 88) and ((vector_no) <= 95) then
        WREG = PIR11 and byte(1 << ((vector_no) mod 8))
  #endif
  #if (_NUM_INT_VECTORS > 96)
    elseif ((vector_no) >= 96) and ((vector_no) <= 103) then
        WREG = PIR12 and byte(1 << ((vector_no) mod 8))
  #endif
  #if (_NUM_INT_VECTORS > 104)
    elseif ((vector_no) >= 104) and ((vector_no) <= 111) then
        WREG = PIR13 and byte(1 << ((vector_no) mod 8))
  #endif
  #if (_NUM_INT_VECTORS > 112)
    elseif ((vector_no) >= 112) and ((vector_no) <= 119) then
        WREG = PIR14 and byte(1 << ((vector_no) mod 8))
  #endif
  #if (_NUM_INT_VECTORS > 120)
    elseif ((vector_no) >= 120) and ((vector_no) <= 127) then
        WREG = PIR15 and byte(1 << ((vector_no) mod 8))
  #endif
    endif    
end macro

// this function can be used to test the result of INTR_REQ()
// returns:
//   true = xxx_IF bit was 1
//   false = xxx_IF bit was 0
// note: this routine relies on the STATUS reg Z flag, so it must be called
//   directly after using INTR_REQ()
// example:
//    INTR_REQ(IRQ_SWINT)             // get intr IF bit (sets WREG and STATUS.Z)
//    if (INTR_FLAG()) then           // if intr IF flag set then
//        INTR_REQ(IRQ_SWINT, 0)      // clear IF
//    endif
public inline function INTR_FLAG() as boolean
    result = false
    if (STATUS.bits(2) = 0) then    // if ZERO flag is clear then the xxx_IF bit was 1
        result = true
    endif        
end function

//
//-----------------------------------------------------------------------------
// INIT_VIC()
//   set VIC peripheral to its poweron settings
// parameters:
//   clear_pirx = 0   do not clear PIRx registers
//   clear_pirx = 1   clear PIRx registers
//-----------------------------------------------------------------------------
//
public macro INIT_VIC(clear_pirx)
    INTCON0 = 0     // disable GIE/GIEL
    
    PIE0 = 0        // disable all individual IE
    PIE1 = 0
    PIE2 = 0
    PIE3 = 0
    PIE4 = 0
    PIE5 = 0
    PIE6 = 0
    PIE7 = 0
    PIE8 = 0
    PIE9 = 0
  #if (_intvectors > 80)
    PIE10 = 0
  #endif
  #if (_intvectors > 88)
    PIE11 = 0
  #endif
  #if (_intvectors > 96)
    PIE12 = 0
  #endif
  #if (_intvectors > 104)
    PIE13 = 0
  #endif
  #if (_intvectors > 112)
    PIE14 = 0
  #endif
  #if (_intvectors > 120)
    PIE15 = 0
  #endif
    
  if (clear_pirx > 0) then
    PIR0 = 0        // clear all individual IF
    PIR1 = 0
    PIR2 = 0
    PIR3 = 0
    PIR4 = 0
    PIR5 = 0
    PIR6 = 0
    PIR7 = 0
    PIR8 = 0
    PIR9 = 0
  #if (_intvectors > 80)
    PIR10 = 0
  #endif
  #if (_intvectors > 88)
    PIR11 = 0
  #endif
  #if (_intvectors > 96)
    PIR12 = 0
  #endif
  #if (_intvectors > 104)
    PIR13 = 0
  #endif
  #if (_intvectors > 112)
    PIR14 = 0
  #endif
  #if (_intvectors > 120)
    PIR15 = 0
  #endif
    endif

    IPR0 = $FF      // set all to high-priority
    IPR1 = $FF
    IPR2 = $FF
    IPR3 = $FF
    IPR4 = $FF
    IPR5 = $FF
    IPR6 = $FF
    IPR7 = $FF
    IPR8 = $FF
    IPR9 = $FF
  #if (_intvectors > 80)
    IPR10 = $FF
  #endif
  #if (_intvectors > 88)
    IPR11 = $FF
  #endif
  #if (_intvectors > 96)
    IPR12 = $FF
  #endif
  #if (_intvectors > 104)
    IPR13 = $FF
  #endif
  #if (_intvectors > 112)
    IPR14 = $FF
  #endif
  #if (_intvectors > 120)
    IPR15 = $FF
  #endif
end macro

//
//-----------------------------------------------------------------------------
// interrupt context
//-----------------------------------------------------------------------------
// NOTE:
// '#option _IVT_SAVE_CONTEXT' must be set to the proper value to allow the module to
// include support for the desired functionality and remove unneeded code and variables 
//
// there are two levels of hardware shadow registers for saving context:
// one for the main context (high-priority) and one for the low-priority isr
//
// in addition to the PC, the xv18 hardware context automatically saves all CPU registers-
//   STATUS, WREG, BSR, FSR0/1/2, PRODL/H and PCLATH/U
// the only important registers it does not save are the TBLPTRU/H/L and TABLAT registers,
// which you should save/restore if the isr accesses const data arrays or strings
//
// there are three different implementations of SAVE_CONTEXT/RESTORE_CONTEXT macros which save the 
// hardware TBLPTRU/H/L/TABLAT registers, and usage depends on the INTR_PRIORITY setting of the ISR...
//      SAVE_CONTEXT()                  // high-priority only 
//      SAVE_CONTEXT(priority)          // high or low... change 'priority' to match ISR INTR_PRIORITY... 0=low, 1=high
//      SAVE_CONTEXT_ISS()              // use run-time Interrupt State Status from INTCON1 (uses more code)
//
// in addition to the TBLPTR registers, you may also need to save/restore the state of the SF system variables
// this is required if the isr uses system library support functions that reference 'SB_SVxx' variables,
// such as multiply or divide
//      SAVE_SYSTEM_CONTEXT()           // high-priority only 
//      SAVE_SYSTEM_CONTEXT(priority)   // high or low... change 'priority' to match ISR INTR_PRIORITY... 0=low, 1=high
//      SAVE_SYSTEM_CONTEXT_ISS()       // use run-time Interrupt State Status from INTCON1 (uses more code)
// these macros take the place of the compiler's equivalent built-in 'save(0)/restore' functions, and 
// if enabled by the _IVT_SAVE_CONTEXT setting will also save the TBLPTRU/H/L and TABLAT registers
//
// SAVE_SYSTEM_CONTEXT allows multiple isr handlers to share the context storage for SF system variables... 
// since isr handlers are defined as events, using 'save(0)' would result in each handler allocating an 
// additional 31 bytes of ram (25 sys regs + 6 FSR0, FSR1, and PROD hdw registers) which is unnecessary. 
// SAVE_SYSTEM_CONTEXT is also slightly faster since it does not have to save the FSRx and PROD registers 
// which are saved automatically as part of the hardware shadow registers.
// the SAVE_xxxx_ISS() versions are for use in rare situations where an isr has a dynamically changing priority
//
// INTCON1[7:6] can be used to identify the current Interrupt State Status (ISS)...
//   11 = High priority ISR executing, high-priority int was received while a low-priority ISR was executing
//   10 = High priority ISR executing, high-priority int was received in main routine
//   01 = Low priority ISR executing, low-priority int was received in main routine
//   00 = Main routine executing
// INTCON1 bit 7 (STAT1) can be used to distinguish between high-priority and low priority
//
// NOTE: 
// along with the above macros, you can also use the compiler's built-in 'save()/restore' functions to save
// the state of any local variables used by subroutines and functions that are called by their IRQ event handler.
// if used, the order of save()/SAVE_SYSTEM_CONTEXT is unimportant, but the order of 'saves' must match
// the order of 'restores' else a system error will occur.
// see the interrupt handler template at the end for an example
//

// data structures for saving TABLEPTR registers and SF SYSTEM variables
#if ((_IVT_SAVE_CONTEXT and _IVTscHPTBL) <> 0)
  public dim hwcontext1(4) as byte            // high-priority TBLPTRU/H/L and TABLAT context
#endif
#if ((_IVT_SAVE_CONTEXT and _IVTscLPTBL) <> 0)
  public dim hwcontext0(4) as byte            // low-priority TBLPTRU/H/L and TABLAT  context
#endif
#if ((_IVT_SAVE_CONTEXT and _IVTscHPSYS) <> 0)
  public dim syscontext1(SB_SV_VARS) as byte  // high-priority system variables context
#endif
#if ((_IVT_SAVE_CONTEXT and _IVTscLPSYS) <> 0)
  public dim syscontext0(SB_SV_VARS) as byte  // low-priority system variables context
#endif

// INTCON1 STAT1 (bit 7) can be used to distinguish between high-priority and low priority
dim ISS_STAT as INTCON1.bits(7)

// SAVE_CONTEXT()
// RESTORE_CONTEXT()
//   save/restore the TBLPTRU/H/L and TABLAT registers
// for use with high-priority only or priority disabled (IPEN=0)
public macro SAVE_CONTEXT()
  if ((ivt_save_context and IVTscHPTBL) = 0) then
    checkparam(etError, "statement requires #option _IVT_SAVE_CONTEXT = _IVTscHPTBL")
  endif
  #if ((_IVT_SAVE_CONTEXT and _IVTscHPTBL) <> 0)
    hwcontext1(3) = TABLAT    // high-priority
    hwcontext1(2) = TBLPTRU
    hwcontext1(1) = TBLPTRH
    hwcontext1(0) = TBLPTRL
  #endif
end macro

public macro RESTORE_CONTEXT()
  if ((ivt_save_context and IVTscHPTBL) = 0) then
    checkparam(etError, "statement requires #option _IVT_SAVE_CONTEXT = _IVTscHPTBL")
  endif
  #if ((_IVT_SAVE_CONTEXT and _IVTscHPTBL) <> 0)
    TABLAT  = hwcontext1(3)   // high-priority
    TBLPTRU = hwcontext1(2)
    TBLPTRH = hwcontext1(1)
    TBLPTRL = hwcontext1(0)
  #endif
end macro

// SAVE_CONTEXT(ipr)
// RESTORE_CONTEXT(ipr)
//   save/restore the TBLPTRU/H/L and TABLAT registers
// parameters:
//   ipr = 0 for low-priority context
//   ipr = 1 for high-priority context
public macro SAVE_CONTEXT(ipr)
  if not checkparam(ipr, cpConst) then
    checkparam(etError, "SAVE_CONTEXT requires const parameter")
  elseif (ipr = 0) and ((ivt_save_context and IVTscLPTBL) = 0) then
    checkparam(etError, "statement requires #option _IVT_SAVE_CONTEXT = _IVTscLPTBL")
  elseif (ipr = 1) and ((ivt_save_context and IVTscHPTBL) = 0) then
    checkparam(etError, "statement requires #option _IVT_SAVE_CONTEXT = _IVTscHPTBL")
  endif
  #if ((_IVT_SAVE_CONTEXT and _IVTscLPTBL) <> 0)
    if (ipr = 0) then
        hwcontext0(3) = TABLAT    // low-priority
        hwcontext0(2) = TBLPTRU
        hwcontext0(1) = TBLPTRH
        hwcontext0(0) = TBLPTRL
    endif
  #endif
  #if ((_IVT_SAVE_CONTEXT and _IVTscHPTBL) <> 0)
    if (ipr = 1) then
        hwcontext1(3) = TABLAT    // high-priority
        hwcontext1(2) = TBLPTRU
        hwcontext1(1) = TBLPTRH
        hwcontext1(0) = TBLPTRL
    endif
  #endif
end macro

public macro RESTORE_CONTEXT(ipr)
  if not checkparam(ipr, cpConst) then
    checkparam(etError, "RESTORE_CONTEXT requires const parameter")
  elseif (ipr = 0) and ((ivt_save_context and IVTscLPTBL) = 0) then
    checkparam(etError, "statement requires #option _IVT_SAVE_CONTEXT = _IVTscLPTBL")
  elseif (ipr = 1) and ((ivt_save_context and IVTscHPTBL) = 0) then
    checkparam(etError, "statement requires #option _IVT_SAVE_CONTEXT = _IVTscHPTBL")
  endif
  #if ((_IVT_SAVE_CONTEXT and _IVTscLPTBL) <> 0)
    if (ipr = 0) then
        TABLAT  = hwcontext0(3)   // low-priority
        TBLPTRU = hwcontext0(2)
        TBLPTRH = hwcontext0(1)
        TBLPTRL = hwcontext0(0)
    endif
  #endif
  #if ((_IVT_SAVE_CONTEXT and _IVTscHPTBL) <> 0)
    if (ipr = 1) then
        TABLAT  = hwcontext1(3)   // high-priority
        TBLPTRU = hwcontext1(2)
        TBLPTRH = hwcontext1(1)
        TBLPTRL = hwcontext1(0)
    endif
  #endif
end macro

// SAVE_CONTEXT_ISS()
// RESTORE_CONTEXT_ISS()
//   save/restore the TBLPTRU/H/L and TABLAT registers based on current priority in INTCON1 
//   Interrupt State Status bits (STAT[1] 0=low, 1=high) 
// note: not recommended for use unless the isr needs to support dynamic priority
#if ((_IVT_SAVE_CONTEXT and _IVTscLPTBL) <> 0) and ((_IVT_SAVE_CONTEXT and _IVTscHPTBL) <> 0)
public inline sub SAVE_CONTEXT_ISS()
    if (ISS_STAT = 0) then              // STAT1=0... low-priority
        SAVE_CONTEXT(0)
    else                                // STAT1=1... high-priority
        SAVE_CONTEXT(1)
    endif
end sub

public inline sub RESTORE_CONTEXT_ISS()
    if (ISS_STAT = 0) then              // STAT1=0... low-priority
        RESTORE_CONTEXT(0)
    else                                // STAT1=1... high-priority
        RESTORE_CONTEXT(1)
    endif
end sub
#endif      // _IVT_SAVE_CONTEXT = _IVTscLPTBL + _IVTscHPTBL

//
// SYSTEM_CONTEXT routines
//   these replace built-in functions 'save(0)/restore'
//
// SAVE_SYSTEM_CONTEXT()
// RESTORE_SYSTEM_CONTEXT()
//   saves TBLPTRU/H/L/TABLAT registers (optional, if enabled) and SF system variables
//   for use with high-priority only or priority disabled (IPEN=0)
public macro SAVE_SYSTEM_CONTEXT()
  if ((ivt_save_context and IVTscHPSYS) = 0) then
    checkparam(etError, "statement requires #option _IVT_SAVE_CONTEXT = _IVTscHPSYS")
  endif
  #if ((_IVT_SAVE_CONTEXT and _IVTscHPSYS) <> 0)    // high-priority
    #if ((_IVT_SAVE_CONTEXT and _IVTscHPTBL) <> 0)  // save TBLPTR registers (if enabled)
      hwcontext1(3) = TABLAT 
      hwcontext1(2) = TBLPTRU
      hwcontext1(1) = TBLPTRH
      hwcontext1(0) = TBLPTRL
    #endif
    FSR1 = addressof(SB_SV0)        // src
    FSR0 = addressof(syscontext1)   // dest
    WREG = SB_SV_VARS               // len
    repeat
      POSTINC0 = POSTINC1
      WREG = WREG - 1
    until (WREG = 0)
  #endif    
end macro

public macro RESTORE_SYSTEM_CONTEXT()
  if ((ivt_save_context and IVTscHPSYS) = 0) then
    checkparam(etError, "statement requires #option _IVT_SAVE_CONTEXT = _IVTscHPSYS")
  endif
  #if ((_IVT_SAVE_CONTEXT and _IVTscHPSYS) <> 0)    // high-priority
    #if ((_IVT_SAVE_CONTEXT and _IVTscHPTBL) <> 0)  // restore TBLPTR registers (if enabled)
      TABLAT  = hwcontext1(3)
      TBLPTRU = hwcontext1(2)
      TBLPTRH = hwcontext1(1)
      TBLPTRL = hwcontext1(0)
    #endif
    FSR1 = addressof(syscontext1)   // src
    FSR0 = addressof(SB_SV0)        // dest
    WREG = SB_SV_VARS               // len
    repeat
      POSTINC0 = POSTINC1
      WREG = WREG - 1
    until (WREG = 0)    
  #endif
end macro

// SAVE_SYSTEM_CONTEXT(ipr)
// RESTORE_SYSTEM_CONTEXT(ipr)
//   save/restore the TBLPTRU/H/L/TABLAT registers and SF system variables
// parameters:
//   ipr = 0 for low-priority context
//   ipr = 1 for high-priority context
public macro SAVE_SYSTEM_CONTEXT(ipr)
  if not checkparam(ipr, cpConst) then
    checkparam(etError, "SAVE_SYSTEM_CONTEXT requires const parameter")
  elseif (ipr = 0) and ((ivt_save_context and IVTscLPSYS) = 0) then
    checkparam(etError, "statement requires #option _IVT_SAVE_CONTEXT = _IVTscLPSYS")
  elseif (ipr = 1) and ((ivt_save_context and IVTscHPSYS) = 0) then
    checkparam(etError, "statement requires #option _IVT_SAVE_CONTEXT = _IVTscHPSYS")
  endif
  #if ((_IVT_SAVE_CONTEXT and _IVTscLPSYS) <> 0)    // low-priority
    if (ipr = 0) then
      #if ((_IVT_SAVE_CONTEXT and _IVTscLPTBL) <> 0)  // save TBLPTR registers (optional, if enabled)
        hwcontext0(3) = TABLAT 
        hwcontext0(2) = TBLPTRU
        hwcontext0(1) = TBLPTRH
        hwcontext0(0) = TBLPTRL
      #endif
      FSR1 = addressof(SB_SV0)        // src
      FSR0 = addressof(syscontext0)   // dest
      WREG = SB_SV_VARS               // len
      repeat
        POSTINC0 = POSTINC1
        WREG = WREG - 1
      until (WREG = 0)    
    endif
  #endif
  #if ((_IVT_SAVE_CONTEXT and _IVTscHPSYS) <> 0)    // high-priority
    if (ipr = 1) then
      #if ((_IVT_SAVE_CONTEXT and _IVTscHPTBL) <> 0)  // save TBLPTR registers (optional, if enabled)
        hwcontext1(3) = TABLAT 
        hwcontext1(2) = TBLPTRU
        hwcontext1(1) = TBLPTRH
        hwcontext1(0) = TBLPTRL
      #endif
      FSR1 = addressof(SB_SV0)        // src
      FSR0 = addressof(syscontext1)   // dest
      WREG = SB_SV_VARS               // len
      repeat
        POSTINC0 = POSTINC1
        WREG = WREG - 1
      until (WREG = 0)    
    endif
  #endif  
end macro

public macro RESTORE_SYSTEM_CONTEXT(ipr)
  if not checkparam(ipr, cpConst) then
    checkparam(etError, "RESTORE_SYSTEM_CONTEXT requires const parameter")
  elseif (ipr = 0) and ((ivt_save_context and IVTscLPSYS) = 0) then
    checkparam(etError, "statement requires #option _IVT_SAVE_CONTEXT = _IVTscLPSYS")
  elseif (ipr = 1) and ((ivt_save_context and IVTscHPSYS) = 0) then
    checkparam(etError, "statement requires #option _IVT_SAVE_CONTEXT = _IVTscHPSYS")
  endif
  #if ((_IVT_SAVE_CONTEXT and _IVTscLPSYS) <> 0)    // low-priority
    if (ipr = 0) then
      #if ((_IVT_SAVE_CONTEXT and _IVTscLPTBL) <> 0)  // restore TBLPTR registers (if enabled)
        TABLAT  = hwcontext0(3)
        TBLPTRU = hwcontext0(2)
        TBLPTRH = hwcontext0(1)
        TBLPTRL = hwcontext0(0)
      #endif
      FSR1 = addressof(syscontext0)   // src
      FSR0 = addressof(SB_SV0)        // dest
      WREG = SB_SV_VARS               // len
      repeat
        POSTINC0 = POSTINC1
        WREG = WREG - 1
      until (WREG = 0)    
    endif
  #endif
  #if ((_IVT_SAVE_CONTEXT and _IVTscHPSYS) <> 0)    // high-priority
    if (ipr = 1) then
      #if ((_IVT_SAVE_CONTEXT and _IVTscHPTBL) <> 0)  // restore TBLPTR registers (if enabled)
        TABLAT  = hwcontext1(3)
        TBLPTRU = hwcontext1(2)
        TBLPTRH = hwcontext1(1)
        TBLPTRL = hwcontext1(0)
      #endif
      FSR1 = addressof(syscontext1)   // src
      FSR0 = addressof(SB_SV0)        // dest
      WREG = SB_SV_VARS               // len
      repeat
        POSTINC0 = POSTINC1
        WREG = WREG - 1
      until (WREG = 0)    
    endif
  #endif  
end macro

// SAVE_SYSTEM_CONTEXT_ISS()
// RESTORE_CONTEXT_ISS()
//   save/restore the SF system variables (and TBLPTRU/H/L/TABLAT registers) based on current priority in INTCON1 
//   Interrupt State Status bits (STAT[1] 0=low, 1=high) 
// note: not recommended for use unless the isr needs to support dynamic priority
#if ((_IVT_SAVE_CONTEXT and _IVTscLPSYS) <> 0) and ((_IVT_SAVE_CONTEXT and _IVTscHPSYS) <> 0)
public inline sub SAVE_SYSTEM_CONTEXT_ISS()
    if (ISS_STAT = 0) then              // STAT1=0... low-priority
        SAVE_SYSTEM_CONTEXT(0)
    else                                // STAT1=1... high-priority
        SAVE_SYSTEM_CONTEXT(1)
    endif
end sub
    
public inline sub RESTORE_SYSTEM_CONTEXT_ISS()
    if (ISS_STAT = 0) then              // STAT1=0... low-priority
        RESTORE_SYSTEM_CONTEXT(0)
    else                                // STAT1=1... high-priority
        RESTORE_SYSTEM_CONTEXT(1)
    endif
end sub
#endif      // _IVT_SAVE_CONTEXT = _IVTscLPSYS + _IVTscHPSYS

//-----------------------------------------------------------------------------
// interrupt handler return
//-----------------------------------------------------------------------------
// intr return with hardware context restore (PREFERRED)
public inline sub RETFIE_CONTEXT()
  asm-
    RETFIE 1
  end asm
end sub

// intr return (no hardware context restore, not recommended)
// YOU MUST SAVE/RESTORE THE CPU CONTEXT MANUALLY
public inline sub RETFIE_NOCONTEXT()
  asm-
    RETFIE 0
  end asm
end sub

//
//-----------------------------------------------------------------------------
// ENTER_INTR_HANDLER(handler)
//   creates the entry point for an interrupt handler
//   it adds an unmangled ASM label with the interrupt handler event name
//   and sets the BANKSEL register to the access bank
// parameters:
//   handler = name of the interrupt handler (must be an 'event' routine)
//
// **THIS MUST BE THE FIRST EXECUTABLE STATEMENT IN THE EVENT HANDLER**
//  the intr handler event must be located at an addr divisable by 4
//    since the vector table is loaded with 'addr >> 2'
//  if the event routine is not properly aligned then we add a NOP which
//    will move the routine addr label up by 2 bytes, aligning it
//  we account for this +2 alignment in END_IVT which will result
//    in the added NOP being skipped when the vector table entry is set
// 
// because of this restriction, DO NOT declare variables using initializers...
// the initializers will not be executed (ie 'dim b as byte = 1' is NOT allowed) 
// use the old style declarations and initialize them separately after the 
// ENTER_INTR_HANDLER() statement
//-----------------------------------------------------------------------------
//
public macro ENTER_INTR_HANDLER(handler)
  asm-
  .if ($ & 0x3)         ; if not div 4 aligned, add an extra NOP instruction
    nop                 ; to align the label (the NOP will be skipped)
  .endif
handler:                ; new entry point for isr
    movlb ACCESSBANK
  end asm
end macro

//
//-----------------------------------------------------------------------------
// EXIT_INTR_HANDLER()
//   leave an interrupt handler, restoring the hardware context registers (RETFIE 1)
//-----------------------------------------------------------------------------
//
public macro EXIT_INTR_HANDLER()
    RETFIE_CONTEXT()
end macro

//
//-----------------------------------------------------------------------------
// default interrupt handler routine
// CREATE_IVT will initially assign all vectors to point to this routine
// if you wish to override this default routine with your own handler, then:
//  - set '#option _IVT_DEFAULT_HANDLER = false'
//  - define a replacement event 'public event default_interrupt_handler()' routine
//  - add at least one reference to the new routine between CREATE_IVT/END_IVT using
//      SET_IVT_HANDLER(<unused_irq_no>, default_interrupt_handler)
//    if you add it first then it doesn't matter what <irq_no> you use as it will be
//    overwritten. for example:
//      CREATE_IVT(ivt_table)
//      #if (_IVT_DEFAULT_HANDLER = false)
//        SET_IVT_HANDLER(0, default_interrupt_handler)   // assign our local default handler
//      #endif
//      SET_IVT_HANDLER(IRQ_SWINT, vector_0_handler)    // IRQ 0 - IRQ_SWINT
//      END_IVT()
//-----------------------------------------------------------------------------
// 
#if (_IVT_DEFAULT_HANDLER)
public event default_interrupt_handler()
    // define interrupt entry point
    ENTER_INTR_HANDLER(default_interrupt_handler)

    // on entry, WREG contains the interrupt number
    
    // normally you would clear the IF flag, but since we're going 
    // to RESET that isn't required here. TRAP will stop a hardware ICD
  asm-
    TRAP
    RESET
  end asm
    
    // exit interrupt
    EXIT_INTR_HANDLER()
end event

// added v1.10
// if '_IVT_DEFAULT_HANDLER=true' there MUST be a reference to the default_interrupt_handler() event 
// otherwise you'll get IDE 'out of memory' crashes due to the 'ENTER_INTR_HANDLER(default_interrupt_handler)'
// reference but the event code is not included. this dummy access should prevent that from occurring
BEGIN_SB_COMMENT_BLOCK()
REGISTER_INTR_HANDLER(default_interrupt_handler)
END_SB_COMMENT_BLOCK()
#endif      // _IVT_DEFAULT_HANDLER

//
//-----------------------------------------------------------------------------
// module init
//-----------------------------------------------------------------------------
//

// default IPEN setting
#if (_IVT_DEFAULT_IPEN = 0)
  DISABLE_INTR_PRIORITY()
#else
  ENABLE_INTR_PRIORITY()
#endif

#if (_IVT_INIT_VIC)
  INIT_VIC(0)
#endif      // _IVT_INIT_VIC

// module init - these get rid of any 'variable not used' warnings
#if ((_IVT_SAVE_CONTEXT and _IVTscHPTBL) <> 0)
  hwcontext1(0) = 0
#endif
#if ((_IVT_SAVE_CONTEXT and _IVTscLPTBL) <> 0)
  hwcontext0(0) = 0
#endif
#if ((_IVT_SAVE_CONTEXT and _IVTscHPSYS) <> 0)
  syscontext1(0) = 0
#endif
#if ((_IVT_SAVE_CONTEXT and _IVTscLPSYS) <> 0)
  syscontext0(0) = 0
#endif

// added v2.0
// this gets rid of warning 'return value of function INTR_FLAG might be undefined'
BEGIN_SB_COMMENT_BLOCK()
intr_flag()
END_SB_COMMENT_BLOCK()

{
//
//-----------------------------------------------------------------------------
// interrupt handler template
//-----------------------------------------------------------------------------
//
// interrupt handlers must follow the template below:
//  - use 'public event/end event' to define the handler
//  - change 'handler_name' to desired name (2 places)
//      'handler_name' must be a globally unique name
//  - change 'irq_no' to match intr vector number
//
//  - (optional) SAVE_CONTEXT/SAVE_SYSTEM_CONTEXT usage depends on irq INTR_PRIORITY setting...
//      SAVE_CONTEXT()                  // high-priority only 
//      SAVE_CONTEXT(priority)          // high or low... change 'priority' to match irq INTR_PRIORITY... 0=low, 1=high
//      SAVE_CONTEXT_ISS()              // use run-time Interrupt State Status (uses more code)
//  - set '#option _IVT_SAVE_CONTEXT' as approriate
//

// if priorities are used you may want to define a const that can be used wherever
// specifying the irq 'priority' is required
public const handler_name_PRIORITY = HIGH_PRIORITY   // or LOW_PRIORITY

public event handler_name()
    // << variable declarations here... do NOT use initializers >>

    // define interrupt entry point (this MUST be the first executable statement)
    ENTER_INTR_HANDLER(handler_name)        // ** change 'handler_name' to match 'public event' name

    // save context (optional, depends on user code)
    // select one or more of the following:
    'SAVE_CONTEXT()                     // save TABLEPTR registers (for const data array/string access)
    'SAVE_CONTEXT(priority)             // same as above... change 'priority' to match IRQ priority setting
    'SAVE_SYSTEM_CONTEXT()              // saves SF system variables (and TABLEPTR registers if enabled) 
    'SAVE_SYSTEM_CONTEXT(priority)      // same as above... change 'priority' to match IRQ priority setting
    'save(mysub)                        // save local frame variables used by mysub

    // << user code goes here >>
    
    // clear IF interrupt request flag
    INTR_REQ(irq_no, 0)                     // ** change 'irq_no' to match intr vector number
    
    // restore context (if saved above)
    'RESTORE_CONTEXT()                  // restore TABLEPTR registers
    'RESTORE_CONTEXT(priority)          // same as above... change 'priority' to match IRQ priority setting
    'RESTORE_SYSTEM_CONTEXT()           // restores SF system variables (and TABLEPTR registers if enabled) 
    'RESTORE_SYSTEM_CONTEXT(priority)   // same as above... change 'priority' to match IRQ priority setting
    'restore                            // restore local frame variables

    // exit interrupt
    EXIT_INTR_HANDLER()
end event

//----------------------------------------------------------------------------
// example IVT table initialization
//----------------------------------------------------------------------------
// create an IVT and add interrupt handlers to the table
CREATE_IVT(ivt_table)
SET_IVT_HANDLER(IRQ_SWINT, handler_name)
SET_IVT_HANDLER(2, default_interrupt_handler)
END_IVT()
}

end module
